from algoritmia.datastructures.graphs import Digraph


# Ordenación topológica de un digrafo acíclico
# Si el grafo tiene algún ciclo, el algoritmo lanza una excepción
def topological_sort[T](g: Digraph[T]) -> list[T]:
    def traverse_from(v: T, used: set[T]):
        seen.add(v)
        for suc_v in g.succs(v):
            if suc_v in used:  # Cycle detection
                raise Exception("The graph has at least one cycle")
            used.add(suc_v)  # Cycle detection
            if suc_v not in seen:
                traverse_from(suc_v, used)
            used.remove(suc_v)  # Cycle detection
        lv.append(v)

    lv = []
    seen = set()
    for v in g.V:
        if len(g.preds(v)) == 0:
            traverse_from(v, set())
    lv.reverse()
    return lv


if __name__ == '__main__':
    edges = [('C', 'C++'), ('C', 'Java'), ('C', 'Objective-C'), ('C', 'C#'),
             ('C++', 'Java'), ('C++', 'C#'), ('Java', 'C#'),
             ('MT', 'Haskell'), ('Haskell', 'C#')]
    my_graph = Digraph(E=edges)
    print(topological_sort(my_graph))



# coding: latin1
import unittest

from algoritmia.datastructures.linkedlists import LinkedList


class TestLinkedList(unittest.TestCase):
    def setUp(self):
        self.seq = list(list(range(100)) * 2)
        self.ll = LinkedList(self.seq)

    def test_len_onLinkedList_shouldReturnNumberOfItems(self):
        self.assertEqual(len(self.seq), len(self.ll))

    def test_len_onEmptyLinkedList_shouldReturnZero(self):
        self.assertEqual(len(LinkedList()), 0)

    def test_append_newItem_shouldEnlargeIt(self):
        self.seq.append(1000)
        self.ll.append(1000)
        self.assertEqual(len(self.seq), len(self.ll))

    def test_append_newItem_shouldPlaceItInLastPos(self):
        self.seq.append(1000)
        self.ll.append(1000)
        self.assertEqual(self.seq[len(self.seq) - 1], self.ll[len(self.ll) - 1])
        self.assertEqual(self.seq[-1], self.ll[-1])

    def test_insert_itemAtZeroIndex_shouldEnlargeIt(self):
        self.seq[:0] = [1000]
        self.ll.insert(0, 1000)
        self.assertEqual(len(self.seq), len(self.ll))

    def test_insert_itemAtZeroIndex_shouldPlaceItInFirstPos(self):
        self.seq[:0] = [1000]
        self.ll.insert(0, 1000)
        self.assertEqual(self.seq[0], self.ll[0])

    def test_pop_onList_shouldShrinkIt(self):
        self.seq.pop()
        self.ll.pop()
        self.assertEqual(list(self.ll), self.seq)

    def test_pop_onList_shouldShrinkReturnLastItem(self):
        x = self.seq.pop()
        y = self.ll.pop()
        self.assertEqual(x, y)

    def test_pop_onEmptyList_shouldRaiseException(self):
        ll = LinkedList()
        self.assertRaises(IndexError, ll.pop)

    def test_del_firstElementOfList_shouldShrinkIt(self):
        del self.seq[0]
        del self.ll[0]
        self.assertEqual(list(self.ll), self.seq)

    def test_del_firstElementOfEmptyList_shouldShrinkIt(self):
        ll = LinkedList()
        self.assertRaises(IndexError, ll.__delitem__, 0)

    def test_remove_anItemFromList_shouldShrinkIt(self):
        self.ll.remove(10)
        self.seq.remove(10)
        self.assertEqual(list(self.ll), self.seq)

    def test_remove_severalItemsFromList_shouldShrinkIt(self):
        for i in range(5, 20):
            self.ll.remove(i)
            self.seq.remove(i)
            self.assertEqual(list(self.ll), self.seq)

    def test_remove_repeatedItemFromList_shouldEliminateIt(self):
        for i in 10, 19:
            self.ll.remove(i)
            self.seq.remove(i)
            self.ll.remove(i)
            self.seq.remove(i)
        self.assertFalse(19 in self.ll)
        self.assertFalse(10 in self.ll)

    def test_remove_fromEmptyList_shouldRaiseException(self):
        ll = LinkedList()
        self.assertRaises(ValueError, ll.remove, 0)

    def test_remove_aNonExistingItem_shouldRaiseError(self):
        ll = LinkedList()
        self.assertRaises(ValueError, ll.remove, 0)

    def test_iter_shouldYieldAllItsContents(self):
        for i, v in enumerate(self.ll):
            self.assertEqual(v, self.seq[i])

    def test_contains_existingItem_shouldReturnTrue(self):
        self.assertTrue(0 in self.ll)
        self.assertTrue(99 in self.ll)

    def test_contains_nonExistingItem_shouldReturnFalse(self):
        self.assertFalse(1000 in self.ll)
        self.assertFalse(-1 in self.ll)

    def test_getitem_atValidIndex_shouldReturnItsValue(self):
        for i in range(99, -1, -1):
            self.assertEqual(self.ll[i], i)
        for i in range(100):
            self.assertEqual(self.ll[i], i)
        for i in range(len(self.ll)):
            self.assertEqual(self.seq[i], self.ll[i])

    def test_getitem_atValidNegativeIndex_shouldReturnItsValue(self):
        for i in range(-1, -100, -1):
            self.assertEqual(self.ll[i], 100 + i)

        for i in range(len(self.ll)):
            self.assertEqual(self.seq[i], self.ll[i])
        self.assertEqual(self.ll[-len(self.ll)], self.ll[0])

    def test_getitem_atNegativeOrEquivalentPositiveIndex_shouldReturnSameValue(self):
        self.assertEqual(self.ll[-len(self.ll)], self.ll[0])

    def test_setitem_atAnyPosition_shouldSetValue(self):
        for i in range(100):
            self.ll[i] = -i
        for i in range(100):
            self.assertEqual(self.ll[i], -i)

    def test_delitem_atSeveralPlaces_shouldRemoveItems(self):
        for i in range(199, -1, -2):
            del self.ll[i]
        for i in range(100):
            if i % 2 == 1:
                self.assertFalse(i in self.ll)
            else:
                self.assertTrue(i in self.ll)

    def test_delitem_oneByOne_shouldReturnEmptyList(self):
        while len(self.ll) > 0:
            del self.ll[0]
        self.assertRaises(IndexError, self.ll.__delitem__, 0)

    def test_delitem_onEmptyList_shouldRaiseException(self):
        self.assertRaises(IndexError, LinkedList().__delitem__, 0)

    def test_insert_atFirstPos_shouldCreateANewFirstElement(self):
        self.ll.insert(0, 1000)
        self.assertTrue(self.ll[0], 1000)

    def test_insert_atLastPos_shouldCreateANewLastElement(self):
        self.ll.insert(len(self.ll) - 1, 2000)
        self.assertTrue(len(self.ll) - 1, 2000)

    def test_insert_atMidPos_shouldCreateANewMidElement(self):
        self.ll.insert(len(self.ll) // 2, 3000)
        self.assertTrue(len(self.ll) // 2, 3000)

    def test_insert_atInvalidPos_shouldRaiseException(self):
        self.assertRaises(IndexError, self.ll.insert, 10000, 0)

    def test_insert_atNegativePos_shouldCreateANewItemThere(self):
        self.ll.insert(len(self.ll), -10)
        self.assertEqual(self.ll[len(self.ll) - 1], -10)

    def test_len_onList_shouldReturnItsLength(self):
        self.assertEqual(len(self.ll), 200)

    def test_len_onEmptyList_shouldReturnZero(self):
        self.assertEqual(len(LinkedList()), 0)

    def test_repr_ofList_shouldCreateEvaluableString(self):
        self.assertEqual(list(self.ll), list(eval(repr(self.ll))))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_LinkedList']
    unittest.main()



#
# Problema del cambio
#  - La solución óptima es la que utiliza menos monedas
#

from algoritmia.utils import infinity

type Quantity = int  # La cantidad a devolver

type Decision = int  # Un número de de monedas
type Solution = list[Decision]

type Score = int | float  # La puntuación es el número de monedas utilizadas o, si no hay solución, +infinito
type ScoredSolution = tuple[Score, Solution]

type SParams = tuple[Quantity, int]


def solve(Q: Quantity, v: list[int]) -> ScoredSolution:
    def S(q: Quantity, n: int) -> Score:
        # Casos base
        if n == 0 and q == 0: return 0
        if n == 0 and q > 0: return infinity

        # Recursividad con memoización
        if (q, n) not in mem:
            mem[q, n] = infinity, (-1, -1), -1
            for d in range(q // v[n - 1] + 1):
                q_prev, n_prev = q - d * v[n - 1], n - 1
                current_score = S(q_prev, n_prev) + d
                if current_score < mem[q, n][0]:  # Minimización
                    mem[q, n] = current_score, (q_prev, n_prev), d
        return mem[q, n][0]

    mem: dict[SParams, tuple[Score, SParams, Decision]] = {}
    score = S(Q, len(v))
    # Recuperación del camino
    decisions = []
    if score != infinity:
        q, n = Q, len(v)
        while n > 0:
            _, (q, n), d = mem[q, n]
            decisions.append(d)
        decisions.reverse()
    return score, decisions


if __name__ == '__main__':
    Q0, v0 = 11, [1, 2, 5, 10]
    print("Instance:")
    print(f"  Q: {Q0}")
    print(f"  v: {v0}\n")

    print("Scored Solution:", solve(Q0, v0))



from collections.abc import Iterable, Callable

from algoritmia.algorithms.traverse import traverse_bf, traverse_dijkstra_dict, traverse_dijkstra_metric_dict
from algoritmia.datastructures.graphs import IGraph, Digraph, Edge, WeightingFunction
from algoritmia.utils import infinity

# Cinco algoritmos para obtener el camino más corto entre dos vertices de un grafo:
# - Tres basados en recorredores:
#   - shortest_path_unweighted_graph(): Para grafos no ponderados
#   - shortest_path_positive_weighted_graph(): Para grafos ponderados positivos (algoritmo de Dijkstra)
#   - shortest_path_metric_graph(): Para grafos métricos (algoritmo de Dijkstra modificado)
# - Dos que utilizan programación dinamica:
#   - shortest_path_acyclic_digraph(): Para digrafos ponderados acíclicos
#   - shortest_path_digraph()/bellman_ford(): Para digrafos ponderados sin ciclos negativos (algoritmo de Bellman-Ford)

type Path[T] = list[T]  # T es el tipo de los vértices


# ----------------------------------------------------------------
# Tres algoritmos basados en recorredores
# ----------------------------------------------------------------


# Devuelve el camino más corto entre dos vértices en grafos no ponderados
# Coste temporal: O(|V| + |E|)
def shortest_path_unweighted_graph[T](g: IGraph[T], v_source: T, v_target: T) -> Path[T]:
    edges = traverse_bf(g, v_source)
    return path_recover(edges, v_target)


# Devuelve el camino más corto entre dos vértices en grafos ponderados positivos (algoritmo de Dijkstra)
# Coste temporal: O(|V|^2)
def shortest_path_positive_weighted_graph[T](g: IGraph[T], d: WeightingFunction,
                                             v_source: T, v_target: T) -> Path[T]:
    edges = traverse_dijkstra_dict(g, d, v_source)  # Con diccionario: O(|V|^2)
    # edges = traverse_dijkstra_heapmap(g, d, v_source)  # Con diccionario de prioridad: O(|V| + |E| log |V|)
    return path_recover(edges, v_target)


# Devuelve el camino más corto entre dos vértices en grafos métricos (algoritmo de Dijkstra modificado)
# Coste temporal: O(|V|^2)
# En la práctica es O(|V|) para grafos densos y O(sqrt(|V|)) para grafos dispersos.
def shortest_path_metric_graph[T](g: IGraph[T], d: WeightingFunction,
                                  dist: Callable[[T, T], float],
                                  v_source: T, v_target: T) -> Path[T]:
    edges = traverse_dijkstra_metric_dict(g, d, dist, v_source, v_target)
    return path_recover(edges, v_target)


# Dada la lista de aristas que devuelve un recorredor y un vértice final,
# devuelve el camino desde el vértice inicial al final
# Coste temporal: O(|V|), porque los recorredores devuelven una arista por vértice: len(edges) = |V|
def path_recover[T](edges: Iterable[Edge[T]], v: T) -> Path[T]:
    # Creates backpointer dictionary (bp)
    bp = {}
    for o, d in edges:
        bp[d] = o
        if d == v:  # I have all I need
            break
    # Recover the path jumping back
    path = [v]
    while v != bp[v]:
        v = bp[v]
        path.append(v)
    # reverse the path
    path.reverse()
    return path


# ----------------------------------------------------------------
# Dos algoritmos de programación dinámica
# ----------------------------------------------------------------

# En este problema, una decisión es un vértice, por lo tanto, su tipo es T (el genérico del grafo)
# Decision = T
type Solution[T] = list[T]

type Score = int | float
type ScoredSolution[T] = tuple[Score, Solution[T]]

type SParams[T] = T  # Para sp_acyclic_digraph()
type SParams2[T] = tuple[T, int]  # Para sp_digraph()/bellman_ford()


# Devuelve el camino más corto entre dos vértices en dígrafos ponderados acíclicos
# Coste temporal: O(|V| + |E|)
def shortest_path_acyclic_digraph[T](g: Digraph[T],
                                     d: WeightingFunction[T],
                                     v_initial: T,
                                     v_final: T) -> ScoredSolution[T]:
    def S(v: T) -> Score:
        if v == v_initial:
            return 0
        if v not in mem:
            res: list[tuple[Score, SParams]] = []
            for u in g.preds(v):
                c_score: Score = S(u) + d(u, v)
                previous: SParams = u
                # La decision, u, coincide con el previous
                res.append((c_score, previous))
            mem[v] = min(res, default=infinity)
        return mem[v][0]

    mem: dict[SParams, tuple[Score, SParams]] = {}
    score = S(v_final)
    if score == infinity:
        return score, []  # El vértice destino es inalcanzable desde el origen
    v0 = v_final
    sol: list[T] = [v0]
    while v0 != v_initial:
        _, v0 = mem[v0]
        sol.append(v0)
    sol.reverse()
    return score, sol


# Devuelve el camino más corto entre dos vértices en digrafos ponderados sin ciclos negativos
# Utiliza el algoritmo de Bellman-Ford (algoritmo de programación dinámica)
# Coste temporal: O(|V| |E|)
def shortest_path_digraph[T](g: Digraph[T],
                             d: WeightingFunction[T],
                             v_initial: T,
                             v_final: T) -> ScoredSolution[T]:
    def S(v: T, n: int) -> Score:
        if v == v_initial:
            return 0
        if n == 0:
            return infinity
        if (v, n) not in mem:
            res: list[tuple[Score, SParams2]] = []
            for u in g.preds(v):
                c_score: Score = S(u, n - 1) + d(u, v)
                previous: SParams2 = u, n - 1
                # La decision, u, se puede extraer del previous
                res.append((c_score, previous))
            mem[v, n] = min(res, default=infinity)
        return mem[v, n][0]

    mem: dict[SParams2, tuple[Score, SParams2]] = {}
    score: Score = S(v_final, len(g.V) - 1)
    if score == infinity:
        return score, []  # El vértice destino es inalcanzable desde el origen
    v0 = v_final
    n0 = len(g.V) - 1
    sol: list[T] = [v0]
    while v0 != v_initial:
        _, (v0, n0) = mem[v0, n0]
        sol.append(v0)
    sol.reverse()
    return score, sol


bellman_ford = shortest_path_digraph  # Creamos un alias para la función


def example_shortest_path_unweighted_graph():
    # Output: [(0, 0), (0, 1), (1, 1), (1, 2), (0, 2), (0, 3), (1, 3)]
    from algoritmia.datastructures.graphs import UndirectedGraph

    type Vertex = tuple[int, int]

    edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((1, 1), (1, 2)),
             ((2, 0), (2, 1)), ((2, 1), (2, 2)), ((2, 2), (2, 3)), ((0, 1), (1, 1)),
             ((0, 2), (1, 2)), ((0, 3), (1, 3)), ((1, 1), (2, 1)), ((1, 2), (2, 2))]
    g = UndirectedGraph[Vertex](E=edges)
    v_initial0 = (0, 0)
    v_final0 = (1, 3)
    print('shortest_path_unweighted_graph:', shortest_path_unweighted_graph(g, v_initial0, v_final0))


def example_shortest_path_positive_weighted_graph():
    from algoritmia.data.iberia import iberia, km

    print('shortest_path_positive_weighted_graph:',
          shortest_path_positive_weighted_graph(iberia, km, 'Madrid', 'Bilbao'))


def example_shortest_path_metric_graph():
    from algoritmia.data.iberia import iberia, km, coords2d

    def eu_dist0(city_a: str, city_b: str) -> float:
        pos2d_a, pos2d_b = coords2d[city_a], coords2d[city_b]
        dx, dy = pos2d_a[0] - pos2d_b[0], pos2d_a[1] - pos2d_b[1]
        return (dx * dx + dy * dy) ** 0.5

    print('shortest_path_metric_graph:', shortest_path_metric_graph(iberia, km, eu_dist0, 'Madrid', 'Bilbao'))


def example_shortest_path_acyclic_digraph():
    data = {(0, 1): 1, (0, 3): 50, (1, 2): 10, (2, 3): 3, (3, 4): 4}
    g = Digraph(E=data.keys())
    wf = WeightingFunction(data)
    print('shortest_path_acyclic_digraph:', shortest_path_acyclic_digraph(g, wf, 0, 3))


def example_shortest_path_digraph():
    data = {(0, 1): 1, (0, 3): 50, (1, 2): 10, (2, 0): 2,
            (2, 3): -3, (3, 1): -2, (3, 2): 100, (3, 4): 4}
    g = Digraph(E=data.keys())
    wf = WeightingFunction(data)
    print('shortest_path_digraph:', shortest_path_digraph(g, wf, 3, 2))


# ----------------------------------------------------------------
# Programa principal
# ----------------------------------------------------------------


if __name__ == '__main__':
    example_shortest_path_unweighted_graph()
    example_shortest_path_positive_weighted_graph()
    example_shortest_path_metric_graph()
    example_shortest_path_acyclic_digraph()
    example_shortest_path_digraph()



"""
Version: 5.2 (01-dic-2023)
         5.1 (23-nov-2023)
         5.0 (31-oct-2023)
         4.1 (29-sep-2022)
         4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2023
@license: GPL3
"""
from abc import abstractmethod
from functools import total_ordering
from typing import final, Optional, Self

from algoritmia.datastructures.priorityqueues import MaxHeap, MinHeap
from algoritmia.schemes.bt_scheme import DecisionSequence, Score, ScoredSolution
from algoritmia.utils import infinity


# ACERCA DEL TIPO ScoredSolution (importado de bt_scheme)
# Las funciones bab_min_solve y bab_max_solve devuelven Optional[ScoredSolution]:
#   - Devuelven None si no hay solución.
#   - Devuelven la tupla (Score, self.solution()) si hay solución.

# ACERCA DEL CÁLCULO EFICIENTE DE LAS COTAS
# Dado que las cotas son inmutables solo deberían calcularse una vez:
# - Los métodos calculate_opt_bound() y calculate_pes_bound() se llaman sólo en el constructor
#   y sus resultados se guardan en los atributos self._opt y self._pes.
# - Los métodos opt() y pes() sólo devuelven el valor de estos atributos.

# La clase BabDecisionSequence -------------------------------------------------------


@total_ordering  # Implementando < y ==, el resto de operadores de comparación se generan automáticamente
class BabDecisionSequence[TDecision, TExtra](DecisionSequence[TDecision, TExtra]):
    def __init__(self, extra: Optional[TExtra] = None, parent: Self = None, decision: TDecision = None):
        DecisionSequence.__init__(self, extra, parent, decision)
        self._pes = self.calculate_pes_bound()
        self._opt = self.calculate_opt_bound()

    # --- Métodos abstractos nuevos ---

    @abstractmethod
    def calculate_opt_bound(self) -> Score:  # Calcula y devuelve la cota optimista
        pass

    @abstractmethod
    def calculate_pes_bound(self) -> Score:  # Calcula y devuelve la cota pesimista
        pass

    # --- Métodos abstractos heredados  ---

    # @abstractmethod
    # def successors(self) -> Iterator[Self]:
    #     pass

    # @abstractmethod
    # def is_solution(self) -> bool:
    #     pass

    # --- Métodos heredados que se puede sobreescribir en las clases hijas: solution() y state() ---

    # def solution(self) -> Solution:
    #    return self.decisions()

    # def state(self) -> State:
    #    return self.decisions()

    # -- Métodos finales que NO se pueden sobreescribir en las clases hijas ---

    # Cota optimista. Para las soluciones su valor debe coincidir con la puntuación real
    @final
    def opt(self) -> Score:
        return self._opt

    # Cota pesimista. Para las soluciones su valor debe coincidir con la puntuación real
    @final
    def pes(self) -> Score:
        return self._pes

    # Comparar dos BabDecisionSequence es comparar sus cotas optimistas
    @final
    def __lt__(self, other: Self) -> bool:
        return self._opt < other._opt

    @final
    def __eq__(self, other: Self) -> bool:
        return self._opt == other._opt


# Esquemas para BaB --------------------------------------------------------------------------

def bab_min_solve(initial_ds: BabDecisionSequence) -> Optional[ScoredSolution]:
    bps = initial_ds.pes()
    heap = MinHeap([initial_ds])  # máx: MaxHeap
    best_seen = {initial_ds.state(): initial_ds.opt()}
    while len(heap) > 0:
        best_ds = heap.extract_opt()
        if best_ds.is_solution():
            return best_ds.opt(), best_ds.solution()
        for new_ds in best_ds.successors():
            if new_ds.opt() <= bps:  # máx: >=
                bps = min(bps, new_ds.pes())  # máx: max
                new_state = new_ds.state()
                if new_ds.opt() < best_seen.get(new_state, infinity):  # máx: >, -inf
                    best_seen[new_state] = new_ds.opt()
                    heap.add(new_ds)


def bab_max_solve(initial_ds: BabDecisionSequence) -> Optional[ScoredSolution]:
    bps = initial_ds.pes()
    heap = MaxHeap([initial_ds])  # mín: MinHeap
    best_seen = {initial_ds.state(): initial_ds.opt()}
    while len(heap) > 0:
        best_ds = heap.extract_opt()
        if best_ds.is_solution():
            return best_ds.opt(), best_ds.solution()
        for new_ds in best_ds.successors():
            if new_ds.opt() >= bps:  # mín: <=
                bps = max(bps, new_ds.pes())  # mín: min
                new_state = new_ds.state()
                if new_ds.opt() > best_seen.get(new_state, -infinity):  # mín: <, inf
                    best_seen[new_state] = new_ds.opt()
                    heap.add(new_ds)



# Dado el vector de valores y una solución, devuelve el valor de dicha solución
def beneficio(v: list[int], sol: list[float]) -> float:
    return sum([x_i * v[i] for i, x_i in enumerate(sol)])


def mochila_fraq0(C: int, v: list[int], w: list[int]) -> list[float]:
    x: list[float] = [0.0] * len(w)
    for i in range(len(w)):
        x[i] = min(1.0, C / w[i])
        C -= x[i] * w[i]
    return x


def mochila_fraq(C: int, v: list[int], w: list[int]) -> list[float]:
    sorted_indices: list[int] = sorted(range(len(w)), key=lambda i: -v[i] / w[i])
    x: list[float] = [0.0] * len(w)
    for i in sorted_indices:
        x[i] = min(1.0, C / w[i])
        C -= x[i] * w[i]
    return x


if __name__ == '__main__':
    capacity, values, weights = 50, [60, 30, 40, 20, 75], [40, 30, 20, 10, 50]

    solution = mochila_fraq0(capacity, values, weights)
    print(solution, beneficio(values, solution))  # Mal: [1, 0.3333333, 0, 0, 0] 70.0

    solution = mochila_fraq(capacity, values, weights)
    print(solution, beneficio(values, solution))  # Ok: [0.5, 0, 1, 1, 0] 90.0



from collections.abc import Callable, Iterable
from typing import Optional

infinity = float("infinity")


def argmin[T, S](iterable: Iterable[T], fn: Callable[[T], S], ifempty: Optional[T] = None) -> Optional[T]:
    try:
        return min((fn(x), x) for x in iterable)[1]
    except ValueError:
        return ifempty


def argmax[T, S](iterable: Iterable[T], fn: Callable[[T], S], ifempty: Optional[T] = None) -> Optional[T]:
    try:
        return max((fn(x), x) for x in iterable)[1]
    except ValueError:
        return ifempty



type Interval = tuple[int, int]  # (start, end)  both are >=0


# Devuelve una lista con el máximo de actividades posibles sin solapes
def select(activities: list[Interval]) -> list[Interval]:
    # Ordenamos los índices por hora de finalización de la tarea
    sorted_indices = sorted(range(len(activities)), key=lambda i: activities[i][1])

    res: list[Interval] = []
    last_task_end = 0
    for i in sorted_indices:
        start, end = activities[i]
        if last_task_end <= start:
            res.append(activities[i])
            last_task_end = end
    return res


if __name__ == '__main__':
    activities0 = [(1, 5), (1, 2), (3, 6), (4, 7), (6, 7)]
    print(select(activities0))



from algoritmia.datastructures.graphs import UndirectedGraph

edges = [('Castelló', 'Sagunt'), ('Sagunt', 'València')]

ugraph = UndirectedGraph(E=edges)  # Parameter V is optional
print(ugraph.V)  # Output: {'Sagunt', 'València', 'Castelló'}
print(ugraph.succs('Sagunt'))  # Output: {'València', 'Castelló'}



from collections.abc import Iterator
from dataclasses import dataclass
from random import seed, randint
from typing import Optional, Self

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solutions, bt_vc_solutions, max_solution

# Tipos  --------------------------------------------------------------------------

type Decision = int  # 0 o 1
type Score = int  # El valor de la mochila
type Solution = tuple[Decision, ...]


# Esquema básico  --------------------------------------------------------------------------

def knapsack_solutions(weights: list[int],
                       values: list[int],
                       capacity: int) -> Iterator[Solution]:
    @dataclass
    class Extra:
        weight: int = 0

    class KnapsackDS(DecisionSequence[Decision, Extra]):
        def is_solution(self) -> bool:
            return len(self) == len(values)

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(values):
                new_weight = self.extra.weight + weights[n]
                if new_weight <= capacity:
                    yield self.add_decision(1, Extra(new_weight))
                yield self.add_decision(0, self.extra)

        # Podríamos sobreescribir 'solution()' para devolver también el peso de la mochila:
        # def solution(self) -> Solution:
        #    return self.decisions(), self.extra.weight

    initial_ds = KnapsackDS(Extra())
    return bt_solutions(initial_ds)


ScoredSolution = tuple[int, Solution]


def knapsack_best_solution(weights: list[int],
                           values: list[int],
                           capacity: int) -> Optional[ScoredSolution]:
    def f(solution: Solution) -> int:
        return sum(d * values[i] for i, d in enumerate(solution))

    all_solutions: Iterator[Solution] = knapsack_solutions(weights, values, capacity)
    return max_solution(all_solutions, f)


# Esquema con control de visitados  --------------------------------------------------------------------------

def knapsack_vc_solutions(weights: list[int],
                          values: list[int],
                          capacity: int) -> Iterator[Solution]:
    @dataclass
    class Extra:
        weight: int = 0

    class KnapsackDS(DecisionSequence[Decision, Extra]):
        def is_solution(self) -> bool:
            return len(self) == len(values)

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(values):
                new_weight = self.extra.weight + weights[n]
                if new_weight <= capacity:
                    yield self.add_decision(1, Extra(new_weight))
                yield self.add_decision(0, self.extra)

        def state(self):
            return len(self), self.extra.weight

        # Podríamos sobreescribir 'solution()' para devolver también el valor y el peso de la mochila:
        # def solution(self) -> Solution:
        #    return self.extra.value, self.extra.weight, self.decisions()

    initial_ds = KnapsackDS(Extra())
    return bt_vc_solutions(initial_ds)


# Funciones auxiliares para crear instancias  ----------------------------------------------

def sorted_by_dec_ratio(weights: list[int], values: list[int]) -> tuple[list[int], list[int]]:
    idxs: list[int] = sorted(range(len(weights)), key=lambda i: -values[i] / weights[i])
    w_new = [weights[i] for i in idxs]
    v_new = [values[i] for i in idxs]
    return w_new, v_new


# Crea los problemas con los objetos ordenados de mayor a menor beneficio por kilo
def create_knapsack_problem(num_objects: int) -> tuple[list[int], list[int], int]:
    seed(5)
    w_new = [randint(10, 100) for _ in range(num_objects)]
    v_new = [w_new[i] * randint(1, 4) for i in range(num_objects)]
    capacity = int(sum(w_new) * 0.3)
    weights, values = sorted_by_dec_ratio(w_new, v_new)
    return weights, values, capacity


# Programa principal -------------------------------------------------------------------------


if __name__ == "__main__":
    # Cada instancia es una tupla (pesos, valores, capacidad):

    # Solution: value = 354, decisions = (1, 0, 1, 0, 0, 0))
    i1 = create_knapsack_problem(6)

    # Solution: value = 1118, decisions = (1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0)
    i2 = create_knapsack_problem(20)

    for W, V, C in [i1, i2]:
        print("-" * 80)
        print(f"Instancia:\n  Pesos = {W}\n  Valores = {V}\n  Capacidad = {C}\n")
        print(f"knapsack_solutions    - Number of solutions: {len(list(knapsack_solutions(W, V, C)))}")
        print(f"knapsack_solutions    - Best solution: {knapsack_best_solution(W, V, C)}")
        print(f"knapsack_vc_solutions - Number of solutions: {len(list(knapsack_vc_solutions(W, V, C)))}")
    print("-" * 80)



"""
2024-09-20: Version 1.8 # Deja cambiar el title, restaura las celdas marcadas al volver al modo laberinto
2021-09-14: Version 1.7 # cambia las teclas: "esc" o "return" para cerrar
2021-09-13: Version 1.6 # corrige bug de no dibujar path
2020-10-26: Version 1.5 # Añade vista dual laberinto/grafo
            Pulsando 'l': activa/desactiva laberinto
            Pulsando 'g': activa/desactiva grafo
            Pulsando 'barra espaciadora': alterna entre laberinto y grafo
2018-10-04: Version 1.4 # add_marked_cell con parametro fillCell=False
            fillCell == False: Por defecto. La marca es un círculo de borde negro más pequeño que la celda
            fillCell == True: La marca rellena completamente la celda
2018-09-22: Version 1.3 (from typing import *)
2016-09-28: Versión 1.2
2015-10-05: Versión 1.1
2012-10-03: Version 1
@author: David Llorens dllorens@uji.es

Dibuja un laberinto a partir de un grafo no dirigido con las siguientes características:
- Los vértices son tuplas de enteros (fila, columna) que se corresponden con las celdas
  del laberinto. El vértice (0,0) se corresponde en el dibujo con la celda superior
  izquierda del laberinto.
- Sólo puede haber una arista entre dos vértices si se corresponden a celdas vecinas en
  el laberinto: La existencia de una arista significa que entre dichas celdas del laberinto
  NO hay muro.
"""

import tkinter

from easypaint import EasyPaint

from algoritmia.datastructures.graphs import UndirectedGraph

type Vertex = tuple[int, int]


class LabyrinthViewer(EasyPaint):
    def __init__(self,
                 lab: UndirectedGraph[Vertex],
                 size: tuple[int, int] | str = (400, 400),
                 margin: int = 10,
                 wall_width: int = 2,
                 title: str = "Labyrinth Viewer - Press 'SPACE', 'g' or 'l'"):
        EasyPaint.__init__(self)

        self.visible_l = True
        self.visible_g = False
        self.wall_width = wall_width
        self.title = title
        self.r_size = size

        # check 'lab' type
        if not isinstance(lab, UndirectedGraph) or \
                any([not isinstance(p, tuple) or len(p) != 2 or not isinstance(p[0], int) or
                     not isinstance(p[1], int) for p in lab.V]):
            raise TypeError("The labyrinth must be an UndirectedGraph where vertex are tuple[int,int]")

        self.g = lab
        self.paths = []
        self.marked_cells = []
        self.margin = margin
        self.max_col = max(p[1] for p in self.g.V)
        self.max_row = max(p[0] for p in self.g.V)

        self.ip = self.op = None

    def set_input_point(self, pos: Vertex):
        self.ip = pos

    def set_output_point(self, pos: Vertex):
        self.op = pos

    def add_marked_cell(self, cell, color='red'):
        self.marked_cells.append((cell, color))

    def add_path(self, path: list[Vertex], color: str = 'red', offset: int = 0):
        self.paths.append((path, color, offset))

    def _draw_path(self, path: list[Vertex], color: str, offset: int):
        u = path[0]
        mw2 = self.mw / 2 + self.cell_size / 2 + offset
        mh2 = self.mh / 2 + self.cell_size / 2 + offset
        for v in path[1:]:
            self.create_line(mw2 + u[1] * self.cell_size, mh2 + u[0] * self.cell_size, mw2 + v[1] * self.cell_size,
                             mh2 + v[0] * self.cell_size, color,
                             capstyle=tkinter.PROJECTING)
            u = v

    def draw_lab(self):
        mw = self.mw
        mh = self.mh
        color = 'black'
        width = self.wall_width

        # Draw borders
        self.create_rectangle(mw / 2, mh / 2, self.canvas_width - mw / 2, self.canvas_height - mh / 2,
                              color, width=width)

        # Draw I and O on input and output cells
        if self.ip is not None:
            self.create_text(mw / 2 + (self.ip[1] + 0.5) * self.cell_size, mh / 2 + (self.ip[0] + 0.5) * self.cell_size,
                             "I", int(self.cell_size / 1.5), "CENTER", "black")
        if self.op is not None:
            self.create_text(mw / 2 + (self.op[1] + 0.5) * self.cell_size, mh / 2 + (self.op[0] + 0.5) * self.cell_size,
                             "O", int(self.cell_size / 1.5), "CENTER", "black")

        # Draw internal walls
        for r in range(self.max_row + 1):
            for c in range(self.max_col + 1):
                u = r, c
                x = mw / 2 + c * self.cell_size
                y = mh / 2 + r * self.cell_size
                cs1 = self.cell_size
                if r < self.max_row and (u not in self.g.V or (r + 1, c) not in self.g.succs(u)):
                    self.create_line(x, y + cs1, x + cs1, y + cs1, color, width=width, capstyle=tkinter.PROJECTING)
                if c < self.max_col and (u not in self.g.V or (r, c + 1) not in self.g.succs(u)):
                    self.create_line(x + cs1, y, x + cs1, y + cs1, color, width=width, capstyle=tkinter.PROJECTING)

        for path, color, offset in self.paths:
            self._draw_path(path, color, offset)

        # Draw marked cells
        # primero dibuja las que rellenan
        for cell, color in self.marked_cells:
            cx = self.mw / 2 + cell[1] * self.cell_size + self.cell_size / 2
            cy = self.mh / 2 + cell[0] * self.cell_size + self.cell_size / 2
            tt = self.create_filled_rectangle(cx - self.cell_size / 2 + 1, cy - self.cell_size / 2 + 1,
                                              cx + self.cell_size / 2, cy + self.cell_size / 2, color, color)
            self.tag_lower(tt)

    def draw_graph(self):
        mw = self.mw
        mh = self.mh

        y_off = 0  # (self.canvas_height + self.cell_size + self.cell_size / 4) / 2
        h = self.cell_size / 2

        for (ru, cu), (rv, cv) in self.g.E:
            self.create_line(cu * self.cell_size + h + mw / 2, y_off + ru * self.cell_size + h + mh / 2,
                             cv * self.cell_size + h + mw / 2, y_off + rv * self.cell_size + h + mh / 2,
                             color='darkgreen')

        for (ru, cu) in self.g.V:
            self.create_filled_circle(cu * self.cell_size + h + mw / 2,
                                      y_off + ru * self.cell_size + h + mh / 2,
                                      self.cell_size / 8,
                                      color='darkgreen',
                                      fill='palegreen')

    def on_key_press(self, keysym):
        if keysym in ['Escape', 'Return']:
            self.close()
        elif keysym.upper() in ['L', 'G', 'SPACE']:
            self.erase()
            if keysym.upper() == 'L':
                self.visible_l = not self.visible_l
            elif keysym.upper() == 'G':
                self.visible_g = not self.visible_g
            elif keysym == 'space':
                if self.visible_l:
                    self.visible_l = False
                    self.visible_g = True
                else:
                    self.visible_l = True
                    self.visible_g = False
            if self.visible_l:
                self.draw_lab()
            if self.visible_g:
                self.draw_graph()

    def main(self):
        if self.r_size == 'FIT' or self.r_size == 'FULL':
            self.cell_size = 100
            self.canvas_height = (self.max_row + 1) * self.cell_size + 2 * self.margin
            self.canvas_width = (self.max_col + 1) * self.cell_size + 2 * self.margin
        else:  # self.r_size = (width, height)
            self.canvas_width, self.canvas_height = self.r_size
            self.cell_size = min(float((self.canvas_width - self.margin * 2) / (self.max_col + 1)),
                                 float((self.canvas_height - self.margin * 2) / (self.max_row + 1)))

        self.easypaint_configure(title=self.title,
                                 background='white',
                                 size=self.r_size,
                                 coordinates=(0, self.canvas_height, self.canvas_width, 0))

        if self.r_size == 'FIT':
            self.size = 'FIT'  # necesario por bug en EasyPaint

        self.mw = self.canvas_width - self.cell_size * (self.max_col + 1)
        self.mh = self.canvas_height - self.cell_size * (self.max_row + 1)

        self.draw_lab()


if __name__ == '__main__':
    e = [((4, 7), (4, 6)), ((4, 7), (4, 8)), ((1, 3), (0, 3)), ((1, 3), (1, 4)), ((4, 8), (4, 9)), ((3, 0), (2, 0)),
         ((3, 0), (4, 0)), ((2, 8), (2, 7)), ((2, 8), (1, 8)), ((2, 1), (2, 0)), ((2, 1), (2, 2)), ((0, 0), (1, 0)),
         ((1, 6), (1, 5)), ((1, 6), (2, 6)), ((3, 7), (3, 8)), ((3, 7), (3, 6)), ((2, 5), (1, 5)), ((2, 5), (2, 4)),
         ((0, 3), (0, 2)), ((4, 0), (4, 1)), ((1, 2), (0, 2)), ((1, 2), (1, 1)), ((4, 9), (3, 9)), ((3, 3), (2, 3)),
         ((3, 3), (4, 3)), ((2, 9), (3, 9)), ((2, 9), (1, 9)), ((4, 4), (3, 4)), ((4, 4), (4, 3)), ((3, 6), (3, 5)),
         ((2, 2), (3, 2)), ((4, 1), (4, 2)), ((1, 1), (1, 0)), ((0, 1), (0, 2)), ((3, 2), (3, 1)), ((2, 6), (2, 7)),
         ((4, 5), (4, 6)), ((0, 4), (0, 5)), ((0, 4), (1, 4)), ((3, 9), (3, 8)), ((0, 5), (0, 6)), ((0, 7), (0, 6)),
         ((0, 7), (1, 7)), ((4, 2), (4, 3)), ((0, 8), (0, 9)), ((3, 5), (3, 4)), ((1, 8), (1, 7)), ((0, 9), (1, 9)),
         ((2, 3), (2, 4))]

    # Laberinto en forma de grafo no dirigido
    graph: UndirectedGraph[Vertex] = UndirectedGraph(E=e)

    # Crea un LabyrinthViewer pasándole el grafo del laberinto

    # Descomenta una línea:
    lv = LabyrinthViewer(graph, size='FIT', margin=10)
    # lv = LabyrinthViewer(graph, size='FULL', margin=10)
    # lv = LabyrinthViewer(graph, size=(400, 400), margin=10)

    # Opcional: Muestra el símbolo 'I' en la celda de entrada al laberinto
    lv.set_input_point((0, 0))

    # Opcional: Visualiza el símbolo 'O' en la celda de salida del laberinto
    lv.set_output_point((4, 9))

    # Opcional: marca una celda
    lv.add_marked_cell((3, 4), 'pink')

    # Opcional: Visualiza un camino en azul
    example_path = [(0, 0), (1, 0), (1, 1), (1, 2), (0, 2), (0, 3), (1, 3), (1, 4), (0, 4), (0, 5), (0, 6),
                    (0, 7), (1, 7), (1, 8), (2, 8), (2, 7), (2, 6), (1, 6), (1, 5), (2, 5), (2, 4), (2, 3),
                    (3, 3), (4, 3), (4, 4), (3, 4), (3, 5), (3, 6), (3, 7), (3, 8), (3, 9), (4, 9)]
    lv.add_path(example_path, 'red', 1)

    # Obligatorio: Muestra el laberinto
    lv.run()



import sys

TVERSION = (3, 1, 2)  # Recuerda cambiarlo también en 'pyproject.toml' <--------- !!!!!
VERSION = '.'.join([str(e) for e in TVERSION])

python_version = sys.version_info

if python_version[0:2] < (3, 12):
    raise ImportError(f"Library 'algoritmia' ({VERSION}) requires Python 3.12 or higher")



def quicksort_basic(v: list[int]) -> list[int]:
    if len(v) <= 1:
        return v
    else:
        pivot = v[0]
        left = [x for x in v if x < pivot]
        right = [x for x in v[1:] if x >= pivot]
        return quicksort_basic(left) + [pivot] + quicksort_basic(right)


def quicksort(v: list[int]):
    def partition(p: int, r: int) -> int:
        pivot, i = v[r - 1], p - 1
        for j in range(p, r - 1):
            if v[j] <= pivot:
                i += 1
                v[i], v[j] = v[j], v[i]
        v[i + 1], v[r - 1] = v[r - 1], v[i + 1]
        return i + 1

    def rec(p: int, r: int):
        if r - p > 1:
            pivot_index = partition(p, r)
            rec(p, pivot_index)
            rec(pivot_index + 1, r)

    rec(0, len(v))


if __name__ == "__main__":
    v0 = [11, 21, 3, 1, 98, 0, 12, 82, 29, 30, 11, 18, 43, 4, 75, 37]
    v1 = v0[:]

    print('quicksort_basic:', quicksort_basic(v0))

    quicksort(v1)
    print('quicksort:', v1)



from abc import abstractmethod, ABC
from collections.abc import Iterable, Sized
from collections import deque

# Fifo() utiliza una 'deque' de Python para tener coste O(1) en el pop()
# Lifo() utiliza una 'list' de Python


class IQueue[T](ABC, Sized):
    def __init__(self, data: deque[T] | list[T] = ()):
        self._list = data

    @abstractmethod
    def push(self, item: T): pass

    @abstractmethod
    def pop(self) -> T: pass

    @abstractmethod
    def top(self) -> T: pass

    def __len__(self) -> int:
        return len(self._list)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({list(self._list)})'


class Fifo[T](IQueue[T]):
    def __init__(self, data: Iterable[T] = ()):
        super().__init__(deque(data))

    def push(self, item: T):
        self._list.append(item)

    def pop(self) -> T:
        return self._list.popleft()

    def top(self) -> T:
        return self._list[0]


class Lifo[T](IQueue[T]):
    def __init__(self, data: Iterable[T] = ()):
        super().__init__(list(data))

    def push(self, item: T):
        self._list.append(item)

    def pop(self) -> T:
        return self._list.pop()

    def top(self) -> T:
        return self._list[-1]


if __name__ == '__main__':
    for q in [Fifo(), Lifo()]:
        for i in range(3):
            q.push(i)
        print("q:", q)
        print("q.top():", q.top())
        print("len(q):", len(q))
        for i in range(3):
            print("q.pop():", q.pop())



from collections.abc import Iterator
from typing import Optional, Self

from algoritmia.datastructures.graphs import UndirectedGraph, WeightingFunction, IGraph
from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solutions, min_solution

# Tipos  --------------------------------------------------------------------------

# En este problema, una decisión es un vértice, por lo tanto, su tipo es T (el genérico del grafo)
# Decision = T

# 'bt_solutions' y 'bt_vc_solutions' devuelven un Iterator del tipo devuelto por el método 'solution' de
# la clase 'DecisionSequence', cuya implementación por defecto devuelve una tupla con las decisiones:
type Solution[T] = tuple[T, ...]


# --------------------------------------------------------------------------------

def hamiltoniancycle_solutions[T](g: IGraph[T]) -> Iterator[Solution]:
    class HamiltonianCycleDS(DecisionSequence[T, None]):  # Como no hay Extra -> ponemos None
        def is_solution(self) -> bool:
            return len(self) == len(g.V) and v_initial in g.succs(self.decision)

        def successors(self) -> Iterator[Self]:
            if len(self) < len(g.V):
                ds_set = set(self.decisions())  # O(|V|)
                for v in g.succs(self.decision):
                    if v not in ds_set:  # O(1), |V| veces
                        yield self.add_decision(v)

    initial_ds = HamiltonianCycleDS()
    v_initial = next(iter(g.V))  # Elegimos un vértice cualquiera como inicial
    initial_ds2 = initial_ds.add_decision(v_initial)  # Forzamos la primera decisión
    return bt_solutions(initial_ds2)


Score = int | float  # La longitud (suma de pesos) del ciclo
ScoredSolution = tuple[Score, Solution]


def hamiltoniancycle_best_solution[T](g: IGraph[T],
                                      wf: WeightingFunction[T]) -> Optional[ScoredSolution]:
    def f(sol: Solution) -> Score:
        return sum(wf(sol[i - 1], sol[i]) for i in range(len(sol)))

    return min_solution(hamiltoniancycle_solutions(g), f)


# Programa principal --------------------------------------------------------------------------------

if __name__ == "__main__":
    edges = [(0, 2), (0, 3), (0, 9), (1, 3), (1, 4), (1, 8), (2, 3), (2, 5), (3, 4),
             (3, 6), (4, 7), (5, 6), (5, 8), (6, 7), (6, 8), (6, 9)]
    g0 = UndirectedGraph(E=edges)
    d0 = dict(((u2, v2), abs(u2 - v2)) for (u2, v2) in g0.E)
    wf0 = WeightingFunction(d0, symmetrical=True)

    print('Basic versión (all solutions):')
    has_solutions = False
    for sol0 in hamiltoniancycle_solutions(g0):
        has_solutions = True
        print(f'\tSolution: {sol0}')
    if not has_solutions:
        print('\tThere are no solutions')

    print('Basic versión (best solution from all solutions):')
    print(f'\tBest solution: {hamiltoniancycle_best_solution(g0, wf0)}')



import colorsys
import random
import sys

from algoritmia.algorithms.traverse import traverse_bf, traverse_df
from algoritmia.datastructures.graphs import UndirectedGraph
from algoritmia.datastructures.mergefindsets import MergeFindSet
from algoritmia.datastructures.queues import Fifo
from algoritmia.viewers.labyrinth_viewer_color import LabyrinthViewerColor

type Vertex = tuple[int, int]


def create_labyrinth_nowalls(num_rows: int, num_cols: int) -> UndirectedGraph[Vertex]:
    # Crear lista de vértices
    v = [(r, c) for r in range(num_rows) for c in range(num_cols)]

    # Crear lista de aristas entre vértices vecinos
    w = []
    for (r, c) in v:
        if r + 1 < num_rows: w.append(((r, c), (r + 1, c)))
        if c + 1 < num_cols: w.append(((r, c), (r, c + 1)))

    return UndirectedGraph(E=w)


def create_labyrinth_mfset(num_rows: int, num_cols: int) -> UndirectedGraph[Vertex]:
    # Crear lista de vértices
    v = [(r, c) for r in range(num_rows) for c in range(num_cols)]

    # Crear lista de aristas entre vértices vecinos
    w = []
    for (r, c) in v:
        if r + 1 < num_rows: w.append(((r, c), (r + 1, c)))
        if c + 1 < num_cols: w.append(((r, c), (r, c + 1)))

    # Barajar la lista de aristas entre vértices vecinos
    random.shuffle(w)
    e = []

    # Crear un MFSet vacío
    mfs = MergeFindSet()

    # AÃ±adir los elementos de 'v' al MFSet
    for vertice in v: mfs.add(vertice)

    # Recorrer aristas en 'w',
    # obtener 'find' de los vertices de cada arista (v1,v2),
    # si 'v1' y 'v2' están en clases diferentes: hacer 'merge' y aÃ±adir arista a 'e'
    for (v1, v2) in w:
        if mfs.find(v1) != mfs.find(v2):
            mfs.merge(v1, v2)
            e.append((v1, v2))

    return UndirectedGraph(E=e)


def matriz_distancias_anchura(grafo, v_inicial):
    dist = {}
    queue = Fifo()
    seen = set()
    queue.push((v_inicial, v_inicial))
    seen.add(v_inicial)
    dist[v_inicial] = 0
    while len(queue) > 0:
        u, v = queue.pop()
        dist[v] = dist[u] + 1
        for suc in grafo.succs(v):
            if suc not in seen:
                seen.add(suc)
                queue.push((v, suc))
    return dist


def int2col(n: int, maxint: int = 255) -> str:
    (r, g, b) = colorsys.hsv_to_rgb(float(n) / maxint * (300 / 360.0), 1.0, 1.0)
    return "#{:02x}{:02x}{:02x}".format(int(255 * r), int(255 * g), int(255 * b))


def int2gray(n: int, maxint: int = 255) -> str:
    k = 100
    m, M = k / 2, 255 - k / 2
    c = 255 - int(float(n) / maxint * (M - m) + m)
    return "#{:02x}{:02x}{:02x}".format(c, c, c)


def create_laberinth(num_rows: int, num_cols: int, tipo: int) -> UndirectedGraph[Vertex]:
    if tipo == 1:
        g = create_labyrinth_mfset(num_rows, num_cols)
    elif tipo == 2:
        g_aux = create_labyrinth_nowalls(num_rows, num_cols)
        edges = list(traverse_bf(g_aux, (0, 0)))[1:]  # quitamos arista fantasma
        g = UndirectedGraph(E=edges)
    elif tipo == 3:
        g_aux = create_labyrinth_nowalls(num_rows, num_cols)
        edges = list(traverse_df(g_aux, (0, 0)))[1:]  # quitamos arista fantasma
        g = UndirectedGraph(E=edges)
    else:
        raise Exception('Unknown tipo')
    return g


# int2col = int2gray


if __name__ == '__main__':
    random.seed(1)
    sys.setrecursionlimit(10000)

    cell_size0 = 13
    num_rows0, num_cols0 = 60, 80

    # Tipo: 1 - MFSet random, 2 - Anchura en lab sin paredes, 3 - Profundidad en lab sin paredes
    tipo0 = 1
    g0 = create_laberinth(num_rows0, num_cols0, tipo0)
    lv0 = LabyrinthViewerColor(g0,
                               canvas_width=num_cols0 * cell_size0 + 20,
                               canvas_height=num_rows0 * cell_size0 + 20, margin=10,
                               wall_width=4)

    start0 = (0, 0)  # (num_rows//2, num_cols//2)
    matriz_dist0 = matriz_distancias_anchura(g0, start0)
    maxvalue0 = max(matriz_dist0.values())
    for (v0, k0) in sorted([(v0, k0) for (k0, v0) in matriz_dist0.items()]):
        lv0.add_marked_cell(k0, int2col(v0, maxvalue0))

    lv0.run()



type Solution = list[int]


def mergesort_basic(v: list[int]) -> list[int]:
    def merge(left: list[int], right: list[int]) -> list[int]:
        c = [0] * (len(left) + len(right))
        i, j, k = 0, 0, 0
        while i < len(left) and j < len(right):
            if left[i] < right[j]:
                c[k] = left[i]
                i += 1
            else:
                c[k] = right[j]
                j += 1
            k += 1
        while i < len(left): c[k] = left[i]; i += 1; k += 1
        while j < len(right): c[k] = right[j]; j += 1; k += 1
        return c

    if len(v) <= 1:
        return v
    else:
        h = len(v) // 2
        return merge(mergesort_basic(v[:h]), mergesort_basic(v[h:]))


def mergesort(v: list[int]):
    def merge(start: int, end: int, c: list[int]):
        mid = (start + end) // 2
        i, j, k = start, mid, start
        while i < mid and j < end:
            if v[i] < v[j]:
                c[k] = v[i]
                i += 1
            else:
                c[k] = v[j]
                j += 1
            k += 1
        while i < mid: c[k] = v[i]; i += 1; k += 1
        while j < end: c[k] = v[j]; j += 1; k += 1
        for k in range(start, end): v[k] = c[k]

    def rec(start: int, end: int, c: list[int]):
        if end - start > 1:
            mid = (start + end) // 2
            rec(start, mid, c)
            rec(mid, end, c)
            merge(start, end, c)

    rec(0, len(v), [0] * len(v))


if __name__ == "__main__":
    v0 = [11, 21, 3, 1, 98, 0, 12, 82, 29, 30, 11, 18, 43, 4, 75, 37]
    v1 = v0[:]
    print(mergesort_basic(v0))

    mergesort(v1)
    print(v1)



from collections.abc import Iterator
from typing import Self

from algoritmia.schemes.dac_scheme import IDivideAndConquerProblem, div_solve


class MergesortProblem(IDivideAndConquerProblem):
    def __init__(self, v: list[int]):
        self.v = v

    def is_simple(self) -> bool:
        return len(self.v) <= 1

    def trivial_solution(self) -> list[int]:
        return self.v

    def divide(self) -> Iterator[Self]:
        mid = len(self.v) // 2
        yield MergesortProblem(self.v[:mid])
        yield MergesortProblem(self.v[mid:])

    def combine(self, sols: Iterator[list[int]]) -> list[int]:
        left, right = sols
        c = [0] * (len(left) + len(right))  # Vector auxiliar
        i, j, k = 0, 0, 0
        while i < len(left) and j < len(right):
            if left[i] < right[j]:
                c[k] = left[i]; i += 1
            else:
                c[k] = right[j]; j += 1
            k += 1
        while i < len(left):  c[k] = left[i];  i += 1; k += 1
        while j < len(right): c[k] = right[j]; j += 1; k += 1
        return c


# Programa principal --------------------------------------
v0 = [11, 21, 3, 1, 98, 0, 12, 82, 29, 30, 11, 18, 43, 4, 75, 37]

# Creamos un problema que cumpla la interfaz IDivideAndConquerProblem
ms_problem0 = MergesortProblem(v0)

# Se lo pasamos a la función div_solve(...) que nos devuelve la solución
solution0 = div_solve(ms_problem0)
print(solution0)



"""
Version:  5.3 (09-ene-2024)
          5.2 (01-dic-2023)
          5.0 (31-oct-2023)
          4.1 (29-sep-2022)
          4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2024
@license: GPL3
"""
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Iterator, Callable, Sized
from typing import Any, final, Optional, Self

# Tipos  --------------------------------------------------------------------------

# ACERCA DEL TIPO Solution
# La implementación por defecto devuelve la tupla de decisiones.
# Podemos sobreescribir solution() en la clase hija para devolver otra cosa.
type Solution = Any

# ACERCA DEL TIPO Score
# Es el tipo devuelto por la función objetivo
type Score = int | float

# ACERCA DEL TIPO ScoredSolution
# Las funciones min_solution y max_solution devuelven Optional[ScoredSolution]:
#   - Devuelven None si no hay solución.
#   - Devuelven la tupla (Score, self.solution()) si hay solución.
type ScoredSolution = tuple[Score, Solution]

# ACERCA DEL TIPO State
# La implementación por defecto devuelve la tupla de decisiones.
# Podemos sobreescribir state() en la clase hija para devolver otra cosa.
type State = Any


# La clase DecisionSequence -------------------------------------------------------

class DecisionSequence[TDecision, TExtra](ABC, Sized):
    def __init__(self,
                 extra: Optional[TExtra] = None,
                 parent: Optional[Self] = None,
                 decision: Optional[TDecision] = None):
        self.extra = extra
        self.parent = parent
        self.decision = decision
        self._len = 0 if parent is None else len(parent) + 1

    # --- Métodos abstractos que hay que implementar en las clases hijas ---

    @abstractmethod
    def is_solution(self) -> bool:
        pass

    @abstractmethod
    def successors(self) -> Iterator[Self]:
        pass

    # --- Métodos que se pueden sobreescribir en las clases hijas: solution() y state() ---

    # Por defecto se devuelve la tupla de decisiones
    # El esquema solo llama a esta función si self.is_solution() es True
    def solution(self) -> Solution:
        return self.decisions()

    # Debe devolver siempre un objeto inmutable
    # Por defecto se devuelve la tupla de decisiones
    def state(self) -> State:
        return self.decisions()

    # -- Métodos finales que NO se pueden sobreescribir en las clases hijas ---

    @final
    def add_decision(self, decision: TDecision, extra: TExtra = None) -> Self:
        return self.__class__(extra, self, decision)

    @final
    def decisions(self) -> tuple[TDecision, ...]:  # Es O(n)
        ds = deque()
        p = self
        while p.parent is not None:
            ds.appendleft(p.decision)
            p = p.parent
        return tuple(ds)

    @final
    def __len__(self) -> int:  # len(objeto) devuelve el número de decisiones del objeto
        return self._len


# Esquema para BT básico --------------------------------------------------------------------------

def bt_solutions(ds: DecisionSequence) -> Iterator[Solution]:
    if ds.is_solution():
        yield ds.solution()
    for new_ds in ds.successors():
        yield from bt_solutions(new_ds)


#  Esquema para BT con control de visitados --------------------------------------------------------

def bt_vc_solutions(initial_ds: DecisionSequence) -> Iterator[Solution]:
    def bt(ds: DecisionSequence) -> Iterator[Solution]:
        if ds.is_solution():
            yield ds.solution()
        for new_ds in ds.successors():
            new_state = new_ds.state()
            if new_state not in seen:
                seen.add(new_state)
                yield from bt(new_ds)

    seen = {initial_ds.state()}  # marcamos initial_ds como visto
    return bt(initial_ds)  # Devuelve un iterador de soluciones


#  Mejor solución --------------------------------------------------------


def min_solution(solutions: Iterator[Solution],
                 f: Callable[[Solution], Score]) -> Optional[ScoredSolution]:
    return min(((f(sol), sol) for sol in solutions), default=None)


def max_solution(solutions: Iterator[Solution],
                 f: Callable[[Solution], Score]) -> Optional[ScoredSolution]:
    return max(((f(sol), sol) for sol in solutions), default=None)



"""

Version: 5.2 (01-dic-2023)
         4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2021
@license: GPL3
"""
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any, Self

type Solution = Any


# Divide and conquer -----------------------------------------------------------


class IDivideAndConquerProblem(ABC):
    @abstractmethod
    def is_simple(self) -> bool: pass

    @abstractmethod
    def trivial_solution(self) -> Solution: pass

    @abstractmethod
    def divide(self) -> Iterator[Self]: pass

    @abstractmethod
    def combine(self, solutions: Iterator[Solution]) -> Solution: pass


def div_solve(problem: IDivideAndConquerProblem) -> Solution:
    if problem.is_simple():
        return problem.trivial_solution()
    else:
        subproblems = problem.divide()
        solutions = (div_solve(p) for p in subproblems)
        return problem.combine(solutions)


# Decrease and conquer -----------------------------------------------------------

class IDecreaseAndConquerProblem(ABC):
    @abstractmethod
    def is_simple(self) -> bool: pass

    @abstractmethod
    def trivial_solution(self) -> Solution: pass

    @abstractmethod
    def decrease(self) -> Self: pass

    def process(self, s: Solution) -> Solution:
        return s


def dec_solve(problem: IDecreaseAndConquerProblem) -> Solution:
    if problem.is_simple():
        return problem.trivial_solution()
    else:
        smaller_problem = problem.decrease()
        return problem.process(dec_solve(smaller_problem))


def tail_dec_solve(problem: IDecreaseAndConquerProblem) -> Solution:
    if problem.is_simple():
        return problem.trivial_solution()
    else:
        return tail_dec_solve(problem.decrease())


def iter_dec_solve(problem: IDecreaseAndConquerProblem) -> Solution:
    while not problem.is_simple():
        problem = problem.decrease()
    return problem.trivial_solution()



def power_naif(a: float, n: int) -> float:
    if n == 0:
        return 1
    elif n == 1:
        return a
    elif n % 2 == 0:
        return power_naif(a, n // 2) * power_naif(a, n // 2)
    else:
        return power_naif(a, n // 2) * power_naif(a, n // 2 + 1)


def power(a: float, n: int) -> float:
    if n == 0:
        return 1
    elif n == 1:
        return a
    elif n % 2 == 0:
        return power(a, n // 2) ** 2
    else:
        return a * power(a, n // 2) ** 2


if __name__ == '__main__':
    print(power_naif(45, 16))
    print(power(45, 16))



from collections.abc import Iterator
from typing import Self

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solutions

# Tipos  --------------------------------------------------------------------------

type Decision = int  # Número de fila donde colocar la reina

# 'bt_solutions' devuelve un Iterator del tipo devuelto por el método 'solution' de
# la clase 'DecisionSequence', cuya implementación por defecto devuelve una tupla con las decisiones:
type Solution = tuple[Decision, ...]


# --------------------------------------------------------------------------------


def nqueens_solutions(board_size: int) -> Iterator[Solution]:
    class NQueensDS(DecisionSequence[Decision, None]):  # Como no hay Extra -> ponemos None
        def is_solution(self) -> bool:
            return len(self) == board_size

        def successors(self) -> Iterator[Self]:
            n = len(self)  # Número de decisiones ya tomadas (reinas colocadas)
            if n < board_size:  # Si quedan decisiones por tomar (reinas por colocar)
                for row in range(board_size):
                    if all(r != row and n - j != abs(row - r)
                           for j, r in enumerate(self.decisions())):
                        yield self.add_decision(row)

    initial_ds = NQueensDS()
    return bt_solutions(initial_ds)


# Programa principal -----------------------------------
if __name__ == "__main__":
    board_size0 = 4

    print('n-queens solutions for n=4:')
    has_solutions = False
    for sol in nqueens_solutions(board_size0):
        has_solutions = True
        print(f'\tSolution: {sol}')
    if not has_solutions:
        print('\tThere are no solutions')



from collections.abc import Iterator
from dataclasses import dataclass
from random import random, seed
from typing import Optional, Self

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solutions, bt_vc_solutions, min_solution

# Tipos  --------------------------------------------------------------------------

type Decision = int
type Score = int
type Solution = tuple[Decision, ...]


# --------------------------------------------------------------------------------

def subsetsum_solutions(e: tuple[int, ...], s: int) -> Iterator[Solution]:
    @dataclass
    class Extra:
        acc_sum: int  # acumulated sum

    class SumSetDS(DecisionSequence[Decision, Extra]):
        def is_solution(self) -> bool:
            return len(self) == len(e) and self.extra.acc_sum == s

        def successors(self) -> Iterator[Self]:
            if len(self) < len(e):
                yield self.add_decision(0, self.extra)
                if self.extra.acc_sum + e[len(self)] <= s:
                    acc_sum2 = self.extra.acc_sum + e[len(self)]
                    yield self.add_decision(1, Extra(acc_sum2))

    initial_ds = SumSetDS(Extra(0))
    return bt_solutions(initial_ds)


ScoredSolution = tuple[int, Solution]


def subsetsum_best_solution(e: tuple[int, ...], s: int) -> Optional[ScoredSolution]:
    def f(solution: Solution) -> int:
        return sum(solution)

    return min_solution(subsetsum_solutions(e, s), f)


# --------------------------------------------------------------------------------

def subsetsum_vc_solutions(e: tuple[int, ...], s: int) -> Iterator[Solution]:
    @dataclass
    class Extra:
        acc_sum: int  # accumulated sum

    class SumSetDS(DecisionSequence[Decision, Extra]):
        def is_solution(self) -> bool:
            return len(self) == len(e) and self.extra.acc_sum == s

        def successors(self) -> Iterator[Self]:
            if len(self) < len(e):
                if self.extra.acc_sum + e[len(self)] <= s:
                    acc_sum2 = self.extra.acc_sum + e[len(self)]
                    yield self.add_decision(1, Extra(acc_sum2))
                yield self.add_decision(0, self.extra)

        # Sobreescribimos 'state()'
        def state(self) -> tuple[int, int]:
            return len(self), self.extra.acc_sum

    initial_ds = SumSetDS(Extra(0))
    return bt_vc_solutions(initial_ds)


# --------------------------------------------------------------------------------

def subsetsum_problem(num_elem):
    seed(42)
    elems = tuple(int(random() * 1000) + 1 for _ in range(num_elem))
    return elems, sum(elems) // 4  # sorted(elems, key=lambda x: -x), sum(elems) // 4


# --------------------------------------------------------------------------------

if __name__ == "__main__":
    # Instance
    elements0, target0 = (640, 777, 276, 224, 677, 893, 87, 422, 30, 306), 393
    print(f"Elements: {elements0}")
    print(f"Target: {target0}\n")

    # Basic version
    print("Basic versión (all solutions):")
    sol0 = None
    for sol0 in subsetsum_solutions(elements0, target0):
        print(f"\tSolution: {sol0}")
    if sol0 is None:
        print("\tThere are no solutions")

    print('Basic versión (best solution from all solutions):')
    print(f"\tBest solution: {subsetsum_best_solution(elements0, target0)}")

    # Visited control version
    print('Visited control version:')
    sol0 = None
    for sol0 in subsetsum_vc_solutions(elements0, target0):
        print(f"\tSolution: {sol0}")
    if sol0 is None:
        print("\tThere are no solutions")
